#include "wifi_manager.h"

#include <string.h>
#include <esp_event.h>
#include <esp_log.h>

#include "lwip/err.h"
#include "lwip/apps/sntp.h"

#define DEFAULT_TX_POWER           14*4            //14dB *4 (1 unit = 0,25dB)
#define DEFAULT_BW                 WIFI_BW_HT20
#define DEFAULT_PRIMARY_CHAN       0
#define DEFAULT_SECONDARY_CHAN     WIFI_SECOND_CHAN_NONE


wifi_config_t wifi_config;
bool wifi_is_enable = false;

static wifi_event_cb on_wifi_evt;
static ip_event_cb on_ip_evt;

wifi_config_param_t wifi_base_config;

static const char *TAG = "WIFI-Manager";

esp_err_t wifi_ap_set_tx_power(int8_t new_tx_power)
{
    if (new_tx_power < 40 || new_tx_power > 82)
        return ESP_ERR_INVALID_ARG;

    wifi_base_config.tx_power = new_tx_power;

    return esp_wifi_set_max_tx_power(new_tx_power);
}

esp_err_t wifi_ap_set_bw_and_channel(wifi_bandwidth_t wifi_bw, uint8_t new_primary_channel, wifi_second_chan_t new_secondary_channel)
{
    esp_err_t err_code;

    if(new_primary_channel > 14)
        return ESP_ERR_INVALID_ARG;

    if(wifi_bw == WIFI_BW_HT40 && new_secondary_channel == WIFI_SECOND_CHAN_NONE)
        return ESP_ERR_INVALID_ARG;

    bool wifi_is_en = wifi_is_enable;
    do
    {
        if(!wifi_is_en)
        {
            err_code = esp_wifi_start();
            if(err_code != ESP_OK)
                break;
            wifi_is_enable = true;
        }

        err_code = esp_wifi_deauth_sta(0);
        if(err_code != ESP_OK)
            break;
        ESP_LOGI(TAG, "New bandwidth: %d", wifi_bw);
        err_code = esp_wifi_set_bandwidth(ESP_IF_WIFI_AP, wifi_bw);
        if(err_code != ESP_OK)
        {
            ESP_LOGE(TAG, "Fail to set bandwidth : err_code: %x", err_code);
            break;
        }
        wifi_base_config.bandwidth = wifi_bw;
        ESP_LOGI(TAG, "New channels: primary :%d, secondary: %d", new_primary_channel, new_secondary_channel);
        err_code = esp_wifi_set_channel(new_primary_channel, new_secondary_channel);
        if(err_code != ESP_OK)
        {
            ESP_LOGE(TAG, "Fail to set channel : err_code: %x", err_code);
        }
        wifi_base_config.primary_channel = new_primary_channel;
        wifi_base_config.secondary_channel = new_secondary_channel;
    } while (0);

    if(!wifi_is_en)
    {
        esp_wifi_stop();
        wifi_is_enable = false;
    }

    if(err_code != ESP_OK)
        ESP_LOGE(TAG, "Fail to change wifi channel");

    return ESP_OK;
}

esp_err_t wifi_enable_interface(bool wifi_enable)
{
    if(wifi_is_enable == wifi_enable)
        return ESP_OK;

    esp_err_t err_code;

    if(wifi_enable)
        err_code = esp_wifi_start();
    else
        err_code = esp_wifi_stop();

    if(err_code == ESP_OK)
        wifi_is_enable = wifi_enable;

    return err_code;
}

static void event_handler(void* arg, esp_event_base_t event_base, int32_t event_id, void* event_data)
{
    ESP_LOGI(TAG, "%s - event id:%d", event_base, event_id);
    if((on_wifi_evt != NULL) && (event_base == WIFI_EVENT))
        on_wifi_evt((wifi_event_t) event_id, event_data);
    else if((on_ip_evt != NULL) && (event_base == IP_EVENT))
        on_ip_evt((ip_event_t) event_id, event_data);
}

static void wifi_ap_init(bool init_netif, char* ap_ssid, char* ap_pwd)
{
    if(init_netif)
    {
        // Initialize TCP/IP network interface (should be called only once in application)
        ESP_ERROR_CHECK(esp_netif_init());
        // Create default event loop that running in background
        ESP_ERROR_CHECK(esp_event_loop_create_default());
    }

    esp_netif_create_default_wifi_ap();

    wifi_init_config_t cfg = WIFI_INIT_CONFIG_DEFAULT();
    ESP_ERROR_CHECK(esp_wifi_init(&cfg));
    ESP_ERROR_CHECK(esp_wifi_set_storage(WIFI_STORAGE_RAM));

    ESP_ERROR_CHECK(esp_event_handler_instance_register(WIFI_EVENT, ESP_EVENT_ANY_ID, &event_handler, NULL, NULL));
    ESP_ERROR_CHECK(esp_event_handler_instance_register(IP_EVENT, ESP_EVENT_ANY_ID, &event_handler, NULL, NULL));

    wifi_config.ap.ssid_len = strlen(ap_ssid);
    wifi_config.ap.max_connection = ESP_MAX_STA_CONN;
    wifi_config.ap.authmode = WIFI_AUTH_WPA_WPA2_PSK;

    memcpy(wifi_config.ap.ssid, ap_ssid, 32);
    memcpy(wifi_config.ap.password, ap_pwd, 64);

    if (strlen(ap_pwd) == 0)
    {
        wifi_config.ap.authmode = WIFI_AUTH_OPEN;
    }

    ESP_LOGI(TAG, "Setting WiFi configuration SSID %s...", wifi_config.ap.ssid);
    ESP_ERROR_CHECK(esp_wifi_set_mode(WIFI_MODE_AP));
    ESP_ERROR_CHECK(esp_wifi_set_config(ESP_IF_WIFI_AP, &wifi_config));

    tcpip_adapter_ip_info_t esp_ap_ip;
    tcpip_adapter_get_ip_info(TCPIP_ADAPTER_IF_AP, &esp_ap_ip);

    uint8_t ap_ip4[4];
    for (int i = 0; i < 4; i++)
    {
        ap_ip4[i] = (uint8_t) (esp_ap_ip.ip.addr >> i*8);
    }
    ESP_LOGI(TAG, "ESP_AP_IP = %d.%d.%d.%d", ap_ip4[0], ap_ip4[1], ap_ip4[2], ap_ip4[3]);

    uint8_t val = 0;
    tcpip_adapter_dhcps_option(TCPIP_ADAPTER_OP_SET, TCPIP_ADAPTER_ROUTER_SOLICITATION_ADDRESS, &val, sizeof(dhcps_offer_t)); //Disable Gateway sent
}

void wifi_manager_init(bool init_netif, wifi_event_cb wifi_cb, ip_event_cb ip_cb, char* ap_ssid, char* ap_pwd)
{
    on_wifi_evt = wifi_cb;
    on_ip_evt = ip_cb;

    memcpy(wifi_base_config.ap_ssid, ap_ssid, sizeof(wifi_base_config.ap_ssid));
    memcpy(wifi_base_config.ap_pwd, ap_pwd, sizeof(wifi_base_config.ap_pwd));

    wifi_ap_init(init_netif, ap_ssid, ap_pwd);

    wifi_base_config.bandwidth = DEFAULT_BW;
    wifi_base_config.primary_channel = DEFAULT_PRIMARY_CHAN;
    wifi_base_config.secondary_channel = DEFAULT_SECONDARY_CHAN;

    wifi_ap_set_bw_and_channel(wifi_base_config.bandwidth, wifi_base_config.primary_channel, wifi_base_config.secondary_channel);

    wifi_base_config.tx_power = DEFAULT_TX_POWER;
    wifi_ap_set_tx_power(wifi_base_config.tx_power);

    esp_read_mac(wifi_base_config.ap_bssid, ESP_MAC_WIFI_SOFTAP);
}
